"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRawOptions = void 0;
const assert_1 = __importDefault(require("assert"));
const stats_1 = require("../stats");
const util_1 = require("../util");
const adapter_rule_use_1 = require("./adapter-rule-use");
const getRawOptions = (options, compiler, processResource) => {
    var _a, _b;
    (0, assert_1.default)(!(0, util_1.isNil)(options.context) && !(0, util_1.isNil)(options.devtool) && !(0, util_1.isNil)(options.cache), "context, devtool, cache should not be nil after defaults");
    const devtool = options.devtool === false ? "" : options.devtool;
    let rawEntry = getRawEntry(options.entry);
    return {
        entry: rawEntry,
        entryOrder: Object.keys(rawEntry),
        mode: options.mode,
        target: getRawTarget(options.target),
        context: options.context,
        output: getRawOutput(options.output),
        resolve: getRawResolve(options.resolve),
        module: getRawModule(options.module, {
            compiler,
            devtool,
            context: options.context
        }),
        externals: options.externals
            ? getRawExternals(options.externals)
            : undefined,
        externalsType: options.externalsType === undefined ? "" : options.externalsType,
        externalsPresets: getRawExternalsPresets(options.externalsPresets),
        devtool,
        optimization: getRawOptimization(options.optimization),
        stats: getRawStats(options.stats),
        devServer: {
            hot: (_b = (_a = options.devServer) === null || _a === void 0 ? void 0 : _a.hot) !== null && _b !== void 0 ? _b : false
        },
        snapshot: getRawSnapshotOptions(options.snapshot),
        cache: {
            type: options.cache ? "memory" : "disable",
            // TODO: implement below cache options
            maxGenerations: 0,
            maxAge: 0,
            profile: false,
            buildDependencies: [],
            cacheDirectory: "",
            cacheLocation: "",
            name: "",
            version: ""
        },
        experiments: getRawExperiments(options.experiments),
        node: getRawNode(options.node),
        // TODO: refactor builtins
        builtins: options.builtins
    };
};
exports.getRawOptions = getRawOptions;
function getRawEntry(entry) {
    const raw = {};
    for (const key of Object.keys(entry)) {
        const runtime = entry[key].runtime;
        const chunkLoading = entry[key].chunkLoading;
        raw[key] = {
            import: entry[key].import,
            publicPath: entry[key].publicPath,
            baseUri: entry[key].baseUri,
            runtime: runtime === false ? undefined : runtime,
            chunkLoading: chunkLoading === false ? "false" : chunkLoading,
            asyncChunks: entry[key].asyncChunks,
            filename: entry[key].filename
        };
    }
    return raw;
}
function getRawTarget(target) {
    if (!target) {
        return [];
    }
    if (typeof target === "string") {
        return [target];
    }
    return target;
}
function getRawAlias(alias = {}) {
    const entires = Object.entries(alias).map(([key, value]) => {
        if (Array.isArray(value)) {
            return [key, value];
        }
        else {
            return [key, [value]];
        }
    });
    return Object.fromEntries(entires);
}
function getRawResolveByDependency(byDependency) {
    if (byDependency === undefined)
        return byDependency;
    return Object.fromEntries(Object.entries(byDependency).map(([k, v]) => [k, getRawResolve(v)]));
}
function getRawResolve(resolve) {
    return {
        ...resolve,
        alias: getRawAlias(resolve.alias),
        fallback: getRawAlias(resolve.fallback),
        extensionAlias: getRawAlias(resolve.extensionAlias),
        byDependency: getRawResolveByDependency(resolve.byDependency)
    };
}
function getRawCrossOriginLoading(crossOriginLoading) {
    if (typeof crossOriginLoading === "boolean") {
        return { type: "bool", boolPayload: crossOriginLoading };
    }
    return { type: "string", stringPayload: crossOriginLoading };
}
function getRawOutput(output) {
    const chunkLoading = output.chunkLoading;
    const wasmLoading = output.wasmLoading;
    const workerChunkLoading = output.workerChunkLoading;
    const workerWasmLoading = output.workerWasmLoading;
    return {
        path: output.path,
        publicPath: output.publicPath,
        clean: output.clean,
        assetModuleFilename: output.assetModuleFilename,
        filename: output.filename,
        chunkFormat: output.chunkFormat === false ? "false" : output.chunkFormat,
        chunkFilename: output.chunkFilename,
        chunkLoading: chunkLoading === false ? "false" : chunkLoading,
        crossOriginLoading: getRawCrossOriginLoading(output.crossOriginLoading),
        cssFilename: output.cssFilename,
        cssChunkFilename: output.cssChunkFilename,
        hotUpdateChunkFilename: output.hotUpdateChunkFilename,
        hotUpdateMainFilename: output.hotUpdateMainFilename,
        uniqueName: output.uniqueName,
        chunkLoadingGlobal: output.chunkLoadingGlobal,
        enabledLibraryTypes: output.enabledLibraryTypes,
        library: output.library && getRawLibrary(output.library),
        strictModuleErrorHandling: output.strictModuleErrorHandling,
        globalObject: output.globalObject,
        importFunctionName: output.importFunctionName,
        iife: output.iife,
        module: output.module,
        wasmLoading: wasmLoading === false ? "false" : wasmLoading,
        enabledWasmLoadingTypes: output.enabledWasmLoadingTypes,
        enabledChunkLoadingTypes: output.enabledChunkLoadingTypes,
        webassemblyModuleFilename: output.webassemblyModuleFilename,
        trustedTypes: output.trustedTypes,
        sourceMapFilename: output.sourceMapFilename,
        hashFunction: output.hashFunction,
        hashDigest: output.hashDigest,
        hashDigestLength: output.hashDigestLength,
        hashSalt: output.hashSalt,
        asyncChunks: output.asyncChunks,
        workerChunkLoading: workerChunkLoading === false ? "false" : workerChunkLoading,
        workerWasmLoading: workerWasmLoading === false ? "false" : workerWasmLoading,
        workerPublicPath: output.workerPublicPath
    };
}
function getRawExternalsPresets(presets) {
    var _a, _b, _c, _d, _e, _f;
    return {
        web: (_a = presets.web) !== null && _a !== void 0 ? _a : false,
        node: (_b = presets.node) !== null && _b !== void 0 ? _b : false,
        electron: (_c = presets.electron) !== null && _c !== void 0 ? _c : false,
        electronMain: (_d = presets.electronMain) !== null && _d !== void 0 ? _d : false,
        electronPreload: (_e = presets.electronPreload) !== null && _e !== void 0 ? _e : false,
        electronRenderer: (_f = presets.electronRenderer) !== null && _f !== void 0 ? _f : false
    };
}
function getRawLibrary(library) {
    const { type, name, export: libraryExport, umdNamedDefine, auxiliaryComment } = library;
    return {
        auxiliaryComment: typeof auxiliaryComment === "string"
            ? {
                commonjs: auxiliaryComment,
                commonjs2: auxiliaryComment,
                amd: auxiliaryComment,
                root: auxiliaryComment
            }
            : auxiliaryComment,
        libraryType: type,
        name: name == null
            ? name
            : typeof name === "object" && !Array.isArray(name)
                ? {
                    amd: name.amd,
                    commonjs: name.commonjs,
                    root: Array.isArray(name.root) || name.root == null
                        ? name.root
                        : [name.root]
                }
                : {
                    amd: Array.isArray(name) ? name[0] : name,
                    commonjs: Array.isArray(name) ? name[0] : name,
                    root: Array.isArray(name) || name == null ? name : [name]
                },
        export: Array.isArray(libraryExport) || libraryExport == null
            ? libraryExport
            : [libraryExport],
        umdNamedDefine
    };
}
function getRawModule(module, options) {
    (0, assert_1.default)(!(0, util_1.isNil)(module.defaultRules), "module.defaultRules should not be nil after defaults");
    // "..." in defaultRules will be flatten in `applyModuleDefaults`, and "..." in rules is empty, so it's safe to use `as RuleSetRule[]` at here
    const ruleSet = [
        { rules: module.defaultRules },
        { rules: module.rules }
    ];
    const rules = ruleSet.map((rule, index) => getRawModuleRule(rule, `ruleSet[${index}]`, options));
    return {
        rules,
        parser: getRawParserOptionsByModuleType(module.parser),
        generator: getRawGeneratorOptionsByModuleType(module.generator)
    };
}
const getRawModuleRule = (rule, path, options) => {
    var _a, _b, _c;
    // Rule.loader is a shortcut to Rule.use: [ { loader } ].
    // See: https://webpack.js.org/configuration/module/#ruleloader
    if (rule.loader) {
        rule.use = [
            {
                loader: rule.loader,
                options: rule.options
            }
        ];
    }
    let funcUse;
    if (typeof rule.use === "function") {
        funcUse = (rawContext) => {
            const context = {
                ...rawContext,
                compiler: options.compiler
            };
            const uses = rule.use(context);
            return (0, adapter_rule_use_1.createRawModuleRuleUses)(uses !== null && uses !== void 0 ? uses : [], `${path}.use`, options);
        };
    }
    return {
        test: rule.test ? getRawRuleSetCondition(rule.test) : undefined,
        include: rule.include ? getRawRuleSetCondition(rule.include) : undefined,
        exclude: rule.exclude ? getRawRuleSetCondition(rule.exclude) : undefined,
        issuer: rule.issuer ? getRawRuleSetCondition(rule.issuer) : undefined,
        dependency: rule.dependency
            ? getRawRuleSetCondition(rule.dependency)
            : undefined,
        descriptionData: rule.descriptionData
            ? Object.fromEntries(Object.entries(rule.descriptionData).map(([k, v]) => [
                k,
                getRawRuleSetCondition(v)
            ]))
            : undefined,
        resource: rule.resource ? getRawRuleSetCondition(rule.resource) : undefined,
        resourceQuery: rule.resourceQuery
            ? getRawRuleSetCondition(rule.resourceQuery)
            : undefined,
        resourceFragment: rule.resourceFragment
            ? getRawRuleSetCondition(rule.resourceFragment)
            : undefined,
        scheme: rule.scheme ? getRawRuleSetCondition(rule.scheme) : undefined,
        mimetype: rule.mimetype ? getRawRuleSetCondition(rule.mimetype) : undefined,
        sideEffects: rule.sideEffects,
        use: typeof rule.use === "function"
            ? { type: "function", funcUse }
            : {
                type: "array",
                arrayUse: (0, adapter_rule_use_1.createRawModuleRuleUses)((_a = rule.use) !== null && _a !== void 0 ? _a : [], `${path}.use`, options)
            },
        type: rule.type,
        parser: rule.parser
            ? getRawParserOptions(rule.parser, (_b = rule.type) !== null && _b !== void 0 ? _b : "javascript/auto")
            : undefined,
        generator: rule.generator
            ? getRawGeneratorOptions(rule.generator, (_c = rule.type) !== null && _c !== void 0 ? _c : "javascript/auto")
            : undefined,
        resolve: rule.resolve ? getRawResolve(rule.resolve) : undefined,
        oneOf: rule.oneOf
            ? rule.oneOf.map((rule, index) => getRawModuleRule(rule, `${path}.oneOf[${index}]`, options))
            : undefined,
        rules: rule.rules
            ? rule.rules.map((rule, index) => getRawModuleRule(rule, `${path}.rules[${index}]`, options))
            : undefined,
        enforce: rule.enforce
    };
};
function getRawRuleSetCondition(condition) {
    if (typeof condition === "string") {
        return {
            type: "string",
            stringMatcher: condition
        };
    }
    if (condition instanceof RegExp) {
        return {
            type: "regexp",
            regexpMatcher: condition.source
        };
    }
    if (typeof condition === "function") {
        return {
            type: "function",
            funcMatcher: condition
        };
    }
    if (Array.isArray(condition)) {
        return {
            type: "array",
            arrayMatcher: condition.map(i => getRawRuleSetCondition(i))
        };
    }
    if (typeof condition === "object" && condition !== null) {
        return {
            type: "logical",
            logicalMatcher: [getRawRuleSetLogicalConditions(condition)]
        };
    }
    throw new Error("unreachable: condition should be one of string, RegExp, Array, Object");
}
function getRawRuleSetLogicalConditions(logical) {
    return {
        and: logical.and
            ? logical.and.map(i => getRawRuleSetCondition(i))
            : undefined,
        or: logical.or ? logical.or.map(i => getRawRuleSetCondition(i)) : undefined,
        not: logical.not ? getRawRuleSetCondition(logical.not) : undefined
    };
}
function getRawParserOptionsByModuleType(parser) {
    return Object.fromEntries(Object.entries(parser).map(([k, v]) => [k, getRawParserOptions(v, k)]));
}
function getRawGeneratorOptionsByModuleType(parser) {
    return Object.fromEntries(Object.entries(parser).map(([k, v]) => [k, getRawGeneratorOptions(v, k)]));
}
function getRawParserOptions(parser, type) {
    if (type === "asset") {
        return {
            type: "asset",
            asset: getRawAssetParserOptions(parser)
        };
    }
    return {
        type: "unknown"
    };
}
function getRawAssetParserOptions(parser) {
    return {
        dataUrlCondition: parser.dataUrlCondition
            ? getRawAssetParserDataUrl(parser.dataUrlCondition)
            : undefined
    };
}
function getRawAssetParserDataUrl(dataUrlCondition) {
    if (typeof dataUrlCondition === "object" && dataUrlCondition !== null) {
        return {
            type: "options",
            options: {
                maxSize: dataUrlCondition.maxSize
            }
        };
    }
    throw new Error(`unreachable: AssetParserDataUrl type should be one of "options", but got ${dataUrlCondition}`);
}
function getRawGeneratorOptions(generator, type) {
    if (type === "asset") {
        return {
            type: "asset",
            asset: generator ? getRawAssetGeneratorOptions(generator) : undefined
        };
    }
    if (type === "asset/inline") {
        return {
            type: "asset/inline",
            assetInline: generator
                ? getRawAssetInlineGeneratorOptions(generator)
                : undefined
        };
    }
    if (type === "asset/resource") {
        return {
            type: "asset/resource",
            assetResource: generator
                ? getRawAssetResourceGeneratorOptions(generator)
                : undefined
        };
    }
    return {
        type: "unknown"
    };
}
function getRawAssetGeneratorOptions(options) {
    return {
        ...getRawAssetInlineGeneratorOptions(options),
        ...getRawAssetResourceGeneratorOptions(options)
    };
}
function getRawAssetInlineGeneratorOptions(options) {
    return {
        dataUrl: options.dataUrl
            ? getRawAssetGeneratorDaraUrl(options.dataUrl)
            : undefined
    };
}
function getRawAssetResourceGeneratorOptions(options) {
    return {
        filename: options.filename,
        publicPath: options.publicPath
    };
}
function getRawAssetGeneratorDaraUrl(dataUrl) {
    if (typeof dataUrl === "object" && dataUrl !== null) {
        return {
            type: "options",
            options: {
                encoding: dataUrl.encoding === false ? "false" : dataUrl.encoding,
                mimetype: dataUrl.mimetype
            }
        };
    }
    throw new Error(`unreachable: AssetGeneratorDataUrl type should be one of "options", but got ${dataUrl}`);
}
function getRawExternals(externals) {
    function getRawExternalItem(item) {
        if (typeof item === "string") {
            return { type: "string", stringPayload: item };
        }
        else if (item instanceof RegExp) {
            return { type: "regexp", regexpPayload: item.source };
        }
        else if (typeof item === "function") {
            return {
                type: "function",
                fnPayload: async (ctx) => {
                    return await new Promise((resolve, reject) => {
                        const promise = item(ctx, (err, result, type) => {
                            if (err)
                                reject(err);
                            resolve({
                                result: getRawExternalItemValueFormFnResult(result),
                                external_type: type
                            });
                        });
                        if (promise && promise.then) {
                            promise.then(result => resolve({
                                result: getRawExternalItemValueFormFnResult(result),
                                external_type: undefined
                            }), e => reject(e));
                        }
                    });
                }
            };
        }
        return {
            type: "object",
            objectPayload: Object.fromEntries(Object.entries(item).map(([k, v]) => [k, getRawExternalItemValue(v)]))
        };
    }
    function getRawExternalItemValueFormFnResult(result) {
        return result === undefined ? result : getRawExternalItemValue(result);
    }
    function getRawExternalItemValue(value) {
        if (typeof value === "string") {
            return { type: "string", stringPayload: value };
        }
        else if (typeof value === "boolean") {
            return { type: "bool", boolPayload: value };
        }
        else if (Array.isArray(value)) {
            return {
                type: "array",
                arrayPayload: value
            };
        }
        throw new Error("unreachable");
    }
    if (Array.isArray(externals)) {
        return externals.map(i => getRawExternalItem(i));
    }
    return [getRawExternalItem(externals)];
}
function getRawOptimization(optimization) {
    (0, assert_1.default)(!(0, util_1.isNil)(optimization.moduleIds) &&
        !(0, util_1.isNil)(optimization.removeAvailableModules) &&
        !(0, util_1.isNil)(optimization.removeEmptyChunks) &&
        !(0, util_1.isNil)(optimization.sideEffects) &&
        !(0, util_1.isNil)(optimization.realContentHash), "optimization.moduleIds, optimization.removeAvailableModules, optimization.removeEmptyChunks, optimization.sideEffects, optimization.realContentHash should not be nil after defaults");
    return {
        splitChunks: toRawSplitChunksOptions(optimization.splitChunks),
        moduleIds: optimization.moduleIds,
        removeAvailableModules: optimization.removeAvailableModules,
        removeEmptyChunks: optimization.removeEmptyChunks,
        sideEffects: String(optimization.sideEffects),
        realContentHash: optimization.realContentHash
    };
}
function toRawSplitChunksOptions(sc) {
    if (!sc) {
        return;
    }
    const { name, cacheGroups = {}, ...passThrough } = sc;
    return {
        name: name === false ? undefined : name,
        cacheGroups: Object.fromEntries(Object.entries(cacheGroups)
            .filter(([_key, group]) => group !== false)
            .map(([key, group]) => {
            group = group;
            const { test, name, ...passThrough } = group;
            const rawGroup = {
                test: test === null || test === void 0 ? void 0 : test.source,
                name: name === false ? undefined : name,
                ...passThrough
            };
            return [key, rawGroup];
        })),
        ...passThrough
    };
}
function getRawSnapshotOptions(snapshot) {
    const { resolve, module } = snapshot;
    (0, assert_1.default)(!(0, util_1.isNil)(resolve) && !(0, util_1.isNil)(module));
    const { timestamp: resolveTimestamp, hash: resolveHash } = resolve;
    const { timestamp: moduleTimestamp, hash: moduleHash } = module;
    (0, assert_1.default)(!(0, util_1.isNil)(resolveTimestamp) &&
        !(0, util_1.isNil)(resolveHash) &&
        !(0, util_1.isNil)(moduleTimestamp) &&
        !(0, util_1.isNil)(moduleHash));
    return {
        resolve: {
            timestamp: resolveTimestamp,
            hash: resolveHash
        },
        module: {
            timestamp: moduleTimestamp,
            hash: moduleHash
        }
    };
}
function getRawExperiments(experiments) {
    const { lazyCompilation, incrementalRebuild, asyncWebAssembly, newSplitChunks, css } = experiments;
    (0, assert_1.default)(!(0, util_1.isNil)(lazyCompilation) &&
        !(0, util_1.isNil)(incrementalRebuild) &&
        !(0, util_1.isNil)(asyncWebAssembly) &&
        !(0, util_1.isNil)(newSplitChunks) &&
        !(0, util_1.isNil)(css));
    return {
        lazyCompilation,
        incrementalRebuild: getRawIncrementalRebuild(incrementalRebuild),
        asyncWebAssembly,
        newSplitChunks,
        css
    };
}
function getRawIncrementalRebuild(inc) {
    if (inc === false) {
        return {
            make: false,
            emitAsset: false
        };
    }
    const { make, emitAsset } = inc;
    (0, assert_1.default)(!(0, util_1.isNil)(make) && !(0, util_1.isNil)(emitAsset));
    return {
        make,
        emitAsset
    };
}
function getRawNode(node) {
    if (node === false) {
        return undefined;
    }
    (0, assert_1.default)(!(0, util_1.isNil)(node.__dirname) && !(0, util_1.isNil)(node.global) && !(0, util_1.isNil)(node.__filename));
    return {
        dirname: String(node.__dirname),
        filename: String(node.__filename),
        global: String(node.global)
    };
}
function getRawStats(stats) {
    var _a;
    const statsOptions = (0, stats_1.normalizeStatsPreset)(stats);
    return {
        colors: (_a = statsOptions.colors) !== null && _a !== void 0 ? _a : false
    };
}
